import threading
import time
import os
import pandas as pd
from pyomo.environ import *
import A_S_R as gt
from calc_rms import cargar_datos, calcular_pedido_pendiente

code = '3_2_1'
folder = 'data/'

def list_to_file(l, filename):
    with open(filename, 'w') as f:
        for s in l:
            f.write(s + '\n')
    f.close()

def extract_Var(var):
    vname = str(var).split('[')[0]
    vvalue = str(var).split('[')[1].split(']')[0].replace(',', '\t') + '\t' + str(var.value)
    return vname, vvalue

# Función para resolver modelo MILP
def resolver_modelo(inst):
    opt = SolverFactory('gurobi')
    resultado = opt.solve(inst)
    inst.solutions.store_to(resultado)

    # Imprimir resultados
    if resultado.solver.status == SolverStatus.ok and resultado.solver.termination_condition == TerminationCondition.optimal:
        resultado.write(filename=os.path.join(folder, code, 'output', 'Results.json'), format='json')
        vnameant = ''
        for v in inst.component_data_objects(Var):
            vname, vvalue = extract_Var(v)
            if vname != vnameant:
                if vnameant != '':
                    list_to_file(table, os.path.join(folder, code, 'output', vnameant + '.txt'))
                table = []
                vnameant = vname
            table.append(vvalue)
        if len(table) > 0:
            list_to_file(table, os.path.join(folder, code, 'output', vnameant + '.txt'))

        print("Resultados:")
        print(resultado)
        print('Total cost: ', inst.obj.expr())
    else:
        print('Solver status is not OK or termination condition is not OPTIMAL')

# Variable global para controlar el evento de reprogramación
evento_reprogramacion = threading.Event()


# Función para cargar el estado de las máquinas
def cargar_valor_de_estado(folder, code):
    ruta_Dpm = os.path.join(folder, code, 'input', 'Status_Machine.tab')
    dm = pd.read_csv(ruta_Dpm, sep='\t')
    return dm

# Función para cargar el estado de las materias primas
def cargar_estado_materia_prima(folder, code):
    ruta_mpf = os.path.join(folder, code, 'input', 'MPF.tab')
    df = pd.read_csv(ruta_mpf, sep='\t')
    return df

# Función para detectar eventos (por ejemplo, averías de máquinas o Escasez de materia prima)
def detectar_eventos(folder, code):
    estado_anterior_dpm = cargar_valor_de_estado(folder, code)
    estado_anterior_materia_prima = cargar_estado_materia_prima(folder, code)

    while True:
        estado_actual_dpm = cargar_valor_de_estado(folder, code)
        estado_actual_materia_prima = cargar_estado_materia_prima(folder, code)

        # Función para detectar avería de máquina
        if (estado_actual_dpm != estado_anterior_dpm).any().any():
            if (estado_actual_dpm['Sm'] != 0).any():
                print("Avería detectada, reprogramando...")
                evento_reprogramacion.set()
            else:
                print("Máquina en funcionamiento normal.")
            # Actualizar el estado anterior
            estado_anterior_dpm = estado_actual_dpm.copy()

        # Función para detectar escasez de materias primas
        if (estado_actual_materia_prima != estado_anterior_materia_prima).any().any():
            if (estado_actual_materia_prima['IdPeriod'] != 0).any():
                print("Se detectó un cambio en el estado de materia prima. Reprogramando...")
                evento_reprogramacion.set()
            # Actualizar el estado anterior
            estado_anterior_materia_prima = estado_actual_materia_prima.copy()

        # Tiempo de espera antes de la próxima detección
        time.sleep(1)


# Función para crear una instancia del modelo y cargar de parámetros de entrada
def crear_instancia():
    dp = DataPortal()
    dp.load(filename=os.path.join(folder, code, 'input', 'Products.tab'), set=gt.m.I)
    dp.load(filename=os.path.join(folder, code, 'input', 'Machines.tab'), set=gt.m.J)
    dp.load(filename=os.path.join(folder, code, 'input', 'Periods.tab'), set=gt.m.T)
    dp.load(filename=os.path.join(folder, code, 'input', 'Machine_Data.tab'), param=(gt.m.ca, gt.m.cc), index=gt.m.J)
    dp.load(filename=os.path.join(folder, code, 'input', 'Initial_Inventory_RawMaterial.tab'), param=gt.m.InvIr, index=gt.m.R)
    dp.load(filename=os.path.join(folder, code, 'input', 'Initial_Inventory_Products.tab'), param=gt.m.InvIp, index=gt.m.I)
    dp.load(filename=os.path.join(folder, code, 'input', 'HoldingCost_Product.tab'), param=gt.m.chp, index=gt.m.I)
    dp.load(filename=os.path.join(folder, code, 'input', 'deferring_Cost.tab'), param=gt.m.cd, index=gt.m.I)
    dp.load(filename=os.path.join(folder, code, 'input', 'Demand.tab'), param=gt.m.d, index=(gt.m.I, gt.m.T))
    dp.load(filename=os.path.join(folder, code, 'input', 'Production_Requirements.tab'), param=(gt.m.req, gt.m.cp, gt.m.st),index=(gt.m.I, gt.m.J))
    dp.load(filename=os.path.join(folder, code, 'input', 'BOM.tab'), param=gt.m.ur, index=(gt.m.R, gt.m.I))
    dp.load(filename=os.path.join(folder, code, 'input', 'Status_Machine.tab'), param=gt.m.sm, index=(gt.m.J, gt.m.T))
    dp.load(filename=os.path.join(folder, code, 'input', 'InProcess_Products.tab'), param=gt.m.P, index=(gt.m.I, gt.m.J, gt.m.T))
    dp.load(filename=os.path.join(folder, code, 'input', 'Produced.tab'), param=gt.m.pb, index=(gt.m.I, gt.m.J, gt.m.T))
    inst = gt.m.create_instance(dp, name='A_S_R')
    return inst


# Función para cargar el plan de producción inicial desde un archivo
def cargar_valores_QP(folder, code):
    valores_QP = {}
    filename = os.path.join(folder, code, 'output', 'QP.txt')
    with open(filename, 'r') as f:
        for line in f:
            i, j, t, valor = line.strip().split('\t')
            valores_QP[(int(i), int(j), int(t))] = float(valor)
    return valores_QP

# Función principal para la lógica de programación y reprogramación
def ejecutar_programa():
    # Ruta de archivo del estado de planificación
    status_file_path = os.path.join(folder, code, 'output', 'status.tab')
    # Verificar si el archivo status.tab existe
    if not os.path.exists(status_file_path):
        # Si no existe, crearlo con el valor 0 (no programado)
        with open(status_file_path, 'w') as f:
            f.write('0')

    # Leer el valor del archivo status.tab
    with open(status_file_path, 'r') as f:
        status = f.read().strip()

    if status == '0': # Verificar la existencia de una programación inicial
        # Crear instancia inicial y realizar la programación inicial
        inst = crear_instancia()
        print("Realizando la programación inicial...")
        resolver_modelo(inst)

        # Actualizar el archivo status.tab a 1 (programación inicial existente)
        with open(status_file_path, 'w') as f:
            f.write('1')
    else:
        print("Inspeccionando estado del taller...")
    print("Inspeccionando estado del taller...")


    # Iniciar el hilo para la detección de eventos
    hilo_eventos = threading.Thread(target=detectar_eventos, args=(folder, code))
    hilo_eventos.daemon = True  # Permite que el hilo se cierre al terminar el programa principal
    hilo_eventos.start()

    while True: #Ciclo de reprogramación
        evento_reprogramacion.wait()  # Espera a que se detecte un evento
        evento_reprogramacion.clear()  # Resetea el evento para la próxima detección

        inst = crear_instancia()

        # Cargar los últimos valores calculados del plan de producción
        valores_QP_actual = cargar_valores_QP(folder, code)

        # Lógica para evento materia prima faltante
        materia_prima_faltante, bom = cargar_datos()
        pedido_pendiente = calcular_pedido_pendiente(materia_prima_faltante, bom)
        # Fijar pedidos pendientes a entregar (Demanda diferida)
        for (i, t), valor in pedido_pendiente.items():
            inst.BO[i, t].fix(valor)

        # Lógica ante evento máquina averiada
        # Fijar plan de producción de un producto
        for (i, j, t), valor in valores_QP_actual.items():
            if inst.sm[j, t] == 0:  # Máquina no averiada
                if inst.P[i, j, t] != 0:  # Producción de un producto ha comenzado
                    inst.QP[i, j, t].fix(valor)
                    print(f"Variable QP[{i}, {j}, {t}] fijada en {valor}")

        # Realizar la reprogramación (resolver el modelo de nuevo)
        resolver_modelo(inst)
        print("Inspeccionando estado del taller...")

# Ejecutar el programa
if __name__ == "__main__":
    ejecutar_programa()

